(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_base.ml,v 1.54 2008/02/11 23:05:59 weis Exp $ *)

(* Base definitions for generation functions *)

open Parsetree
open Code
open Relation
open Genr_pattern
open Longident
;;

type binary_fun_exp = exp -> exp -> exp
;;

(* Do we generate maximum sharing ? *)
let set_sharable_target, get_sharable_target =
  let sharable_target = ref false in
  (fun () -> sharable_target := true),
  (fun () -> !sharable_target)
;;

let get_eq_fun_def () =
  if get_sharable_target () then ["%eq"] else ["%equal"]
;;

let get_eq_fun () =
  if get_sharable_target () then "==" else "="
;;

(* equality function *)
let eq_function x y = infix_name x (get_eq_fun ()) y
;;

(* Generation of an equality function for the type [tname]. *)
let genr_eq (tname, tdecl) =
  Pstr_primitive
    ("eq_" ^ tname,
     { pval_type = Otype.type_of_eq tname tdecl.ptype_params;
       pval_prim = get_eq_fun_def ();
     })
;;

let genr_projection (tname, tdecl) trep =
  Pstr_primitive
    ("from_" ^ tname,
     { pval_type = Otype.projection_type tname tdecl.ptype_params trep;
       pval_prim = ["%identity"];
     })
;;

(* Generation of a pattern *)
let genr_pattern p =
  if get_sharable_target ()
  then add_underscores (remove_topconstr p)
  else remove_topconstr p
;;

(* Building the construction function name out of the generator name. *)
let construction_function_name = String.uncapitalize
;;

let prefix_string pref str = Printf.sprintf "%s_%s" pref str
;;

let rec prefix_ident prefix = function
  | Lident ident -> Lident (prefix_string prefix ident)
  | Ldot (modid, ident) -> Ldot (modid, prefix_string prefix ident)
  | Lapply (modid1, modid2) -> Lapply (modid1, prefix_ident prefix modid2)
;;

let rec prefix_construction_function_ident prefix = function
  | Lident gname ->
      Lident (prefix_string prefix (construction_function_name gname))
  | Ldot (modid, gname) ->
      Ldot (modid, prefix_string prefix (construction_function_name gname))
  | Lapply (modid1, modid2) ->
      Lapply (modid1, prefix_construction_function_ident prefix modid2)
;;

let rec construction_function_ident = function
  | Lident gname -> Lident (construction_function_name gname)
  | Ldot (modid, gname) -> Ldot (modid, construction_function_name gname)
  | Lapply (modid1, modid2) ->
    Lapply (modid1, construction_function_ident modid2)
;;

(* Name generation functions from a generator name:
   [prefix_gname p c] is the name "prefix_name",
   where "prefix" is the contents of p,
   and "name" is the lower case version of c. *)
let prefix_construction_function prefix g =
  prefix_construction_function_ident prefix g.pgen_desc
;;

let prefix_construction_function2 prefix g1 g2 =
  prefix_construction_function (Otype.name_of_longident (prefix_construction_function prefix g1)) g2
;;

let construction_function g = construction_function_ident g.pgen_desc
;;

(* Apply the construction function for generator [g]. *)
let construction g =
  let fid = construction_function_ident g.pgen_desc in
    function
      | [] -> Constant fid
      | [x] -> apply1 fid x
      | xs when Check.is_varyadic_gen g -> apply1 fid (List xs)
      | xs -> apply1 fid (Tuple xs)
;;

let construction0 g = construction g []
and construction1 g x = construction g [x]
and construction2 g x y = construction g [x; y]
;;

(* Apply the generator for generator [g]:
   in case of sharing we still need to call the sharing version of the
   generator. *)
let generator g args =
  let gid = g.pgen_desc in
  if get_sharable_target () && args <> []
  then Apply (Constant (Genr_sharing.mk_sharing_function_ident gid), args)
  else match args with
    | [] -> Constant gid
    | [x] -> Generator (gid, x)
    | _ when Check.is_varyadic_gen g -> Generator (gid, List args)
    | _ -> Generator (gid, Tuple args)
;;

let generator0 g = generator g []
and generator1 g x = generator g [x]
and generator2 g x y = generator g [x; y]
;;

(* Generator pattern construction. *)
let pattern g args =
  let gid = g.pgen_desc in
  if get_sharable_target () && args <> []
  then Genr_sharing.generator_underscore (gid, args)
  else match args with
    | [] -> Constant gid
    | [x] -> Generator (gid, x)
    | _ -> Generator (gid, Tuple args)
;;

let pattern0 g = pattern g []
and pattern1 g x = pattern g [x]
and pattern2 g x y = pattern g [x; y]
;;

(* Data structure for the generation of a construction function. *)
type function_definition = {
  fd_generator : Parsetree.generator;
  fd_relations : Parsetree.relations;
  fd_arity : int;
}
;;

(* Functions to select a clause according to the side argument. *)
let clauses_of_side side left right =
  match side with
  | Left -> [left]
  | Right -> [right]
  | Both -> [left; right]
;;

let clauses_left_or_right side left right =
  match side with
  | Left -> [left]
  | Right -> [right]
  | Both -> assert false
;;

let clauses_both_as_left side left right =
  match side with
  | Left | Both -> [left]
  | Right -> [right]
;;

(* Functions to generate clauses on demand. *)
let opt b c cls = if b then c :: cls else cls
and lazy_opt b c cls = if b then c () :: cls else cls
;;

(* Comparison function. *)

type comparisons = binary_fun_exp * binary_fun_exp * binary_fun_exp
;;

let pervasives_compare = Ldot (Lident "Pervasives", "compare")
;;

let inverse f =
  Let (Lident "compare", [x; y],
      apply1 (Lident "-") (apply2 f x y),
      make_Constant "compare")
;;

let comparison_function = function
  | None -> Constant pervasives_compare
  | Some lid -> Constant lid
;;

let comparison_function_in_rels rels =
  let comp_function_ident = Relation.commutative rels in
  comparison_function comp_function_ident
;;


(* Comparison expressions. *)

(* The usual default infix comparisons. *)
(*let default_comparisons =
  let default_lt x y = infix_name x "<" y
  and default_eq x y = infix_name x "=" y
  and default_gt x y = infix_name x ">" y in
  default_lt, default_eq, default_gt
;;*)

(* The standard encodings of comparisons in terms of
   expression compare x y < (or = or >) 0. *)
let standard_comparisons lid =

  let standard_comp inf lid x y =
    infix_name (apply2 lid x y) inf (make_Constant "0") in

  let standard_lt = standard_comp "<"
  and standard_eq = standard_comp "="
  and standard_gt = standard_comp ">" in

  standard_lt lid, standard_eq lid, standard_gt lid
;;

(* Generation of a triple of comparison functions for a given
   commutative operator. *)
let genr_comparisons = function
  (* Commutative generator with no special comparison provided: use the
     standards comparisons. *)
  | None -> standard_comparisons pervasives_compare
  (* Commutative generator with a special comparison provided: use it! *)
  | Some lid -> standard_comparisons lid
;;

(* Basic idents and functions. *)

let map = function
  | Dist_Inverse -> List.rev_map
  | Dist_Direct -> List.map
;;

let imap = function
  | Dist_Inverse -> Longident.Ldot (Lident "List", "rev_map")
  | Dist_Direct -> Longident.Ldot (Lident "List", "map")
;;
