(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_sharing.ml,v 1.57 2008/02/13 08:03:22 weis Exp $ *)

open Asttypes
open Parsetree
open Longident
open Relation
open Code
open Genr_pattern
open Otype
open Pr_otype
open Check
;;

let list_map_i f l =
  let rec loop i accu = function
  | [] -> accu
  | x :: l -> let h = f i x in loop (succ i) (h :: accu) l in
  List.rev (loop 0 [] l)
;;

(* Various names for functions, built from the name of the type. *)

(* The name of the construction function that will share terms
   with head generator gid (Lident gname). *)
let mk_sharing_function_ident gid = Code.prefix_longident "mk_" gid
;;

(* The modules we generate. *)
let hashing_module_ident = prefix_longident "Hashing_"
and sharing_module_ident = prefix_longident "Sharing_"
;;

(* The name of the private module that defines the relational type. *)
let type_private_module_ident = prefix_longident "Type_"
;;

(* The name of the type as defined in its private module. *)
let type_ident_in_private_module tid =
  let tname = name_of_longident tid in
  Ldot (type_private_module_ident tid, tname)
;;

let type_obj_ident = Ldot (Lident "Obj", "t")
;;

(* The functions we generate. *)
let eq_function_ident = prefix_longident "eq_"
and equal_function_ident = prefix_longident "equal_"
and hash_function_ident = prefix_longident "hash_"
and get_hash_function_ident tid = Code.prefix_longident "get_hash_" tid
;;

(* The name of the sharing hash table. *)
let sharing_hash_table_ident = prefix_longident "htbl_"
;;
let initial_sharing_hash_table_size = 1009
;;

(* The name of the system hash function. *)
let default_hash_function_ident = Ldot (Lident "Hashtbl", "hash")
;;

(* Which function to use to hash arguments of generator gname ? *)
let get_hash_function_ident_for_arg_i tid gname i =
  (* The default hash function is the one of the [Hashtbl] module of the
     standard library. *)
  try
    let gi = Check.find_gname gname in
    let gt = List.nth gi.gi_arg_types (i - 1) in
    begin match gt.ptyp_desc with
     | Ptyp_constr ((Lident _ as gtid), _) when gtid = tid ->
       get_hash_function_ident tid
     | _ ->
       default_hash_function_ident end with
  | Not_found -> default_hash_function_ident
;;

(* Which function to use to compare arguments of generator gname ? *)
let get_eq_function_name_for_arg_i tname gname i =
  try
    let gi = Check.find_gname gname in
    let gt = List.nth gi.gi_arg_types (i - 1) in
    begin match gt.ptyp_desc with
     (* Comparison between values of a maximally shared type: use ==. *)
     | Ptyp_constr (Lident s, _) when s = tname -> "=="
     (* FIXME: we should use == for int, char, bool, ... *)
     (* Conservative approach:
        in any other case use the generic = function. *)
     | _ -> "=" end with
  | Not_found -> "="
;;

let get_eq_function_ident_for_arg_i tid gname i =
  let tname = name_of_longident tid in
  Longident.Lident (get_eq_function_name_for_arg_i tname gname i)
;;

(* The type info that would contain the hash key for a value of
   the relational type. *)
let typ_info = Otype.constant_type "info"
;;

let add_info_to_generator_type ((gid, tycl, rels, loc) as cstr) =
  match tycl with
  | [] -> cstr
  | _ -> (gid, typ_info :: tycl, rels, loc)
;;

let add_info_arg_to_type_kind = function
  | Ptype_abstract _relations as x -> x
  | Ptype_variant cstrs ->
      Ptype_variant (List.map add_info_to_generator_type cstrs)
  | Ptype_record _flds as x -> x
;;

let add_info_arg_to_type_declaration td =
  { td with ptype_kind = add_info_arg_to_type_kind td.ptype_kind }
;;

(*let print_typedef ob tname =
  let type_decl = find_tname tname in
  let gt =
    mktyp
     (Ptyp_constr
       (Ldot (Lident type_module, tname),
        List.map var_type type_decl.ptype_params)) in
  let new_type_decl = {
    type_decl with
      ptype_kind = make_public_type_kind type_decl.ptype_kind;
      ptype_manifest = Some gt;
  } in
  Code.structure_item ob (Pstr_type [tname, new_type_decl])
;;*)

let find_typedef_params tid =
  let tname = name_of_longident tid in
  let type_decl = find_tname tname in
  type_decl.ptype_params
;;

let is_typedef_polymorphic tid =
  find_typedef_params tid != []
;;

let print_type_params ob = function
  | [] -> ()
  | [ x ] -> Printf.bprintf ob "'%s " x
  | x :: l ->
    let print_param ob x = Printf.bprintf ob ", '%s" x in
    let print_params ob l = List.iter (print_param ob) l in
    Printf.bprintf ob "('%s%a) " x print_params l
;;

let longident = Pr_otype.blongident
;;

let print_type_shared ob tid =
  let ty_vars = find_typedef_params tid in
  Printf.bprintf ob "%a%a"
    print_type_params ty_vars
    longident (type_ident_in_private_module tid)
;;

let print_hashing_module ob tid =
  if not (is_typedef_polymorphic tid) then begin
    (* The monomorphic case. *)
    (* We output the module that defines the hash function. *)
    Printf.bprintf ob "\
      \nmodule %a = struct\
      \n  type t = %a;;\
      \n  let equal = %a;;\
      \n  let hash = %a;;\
      \nend;;\n"
      longident (hashing_module_ident tid)
      longident (type_ident_in_private_module tid)
      longident (equal_function_ident tid)
      longident (hash_function_ident tid);
    (* We output the module that defines the sharing hash table. *)
    Printf.bprintf ob "\
      \nmodule %a = Weak.Make (%a);;\n"
      longident (sharing_module_ident tid)
      longident (hashing_module_ident tid);
  end else begin
    (* The polymorphic case. *)
    let print_equal_function_body ob fid =
      Printf.bprintf ob
        "fun x y -> %a (Obj.magic x) (Obj.magic y)"
        longident fid in
    let print_hash_function_body ob fid =
      Printf.bprintf ob
        "fun x -> %a (Obj.magic x : %a)"
        longident fid
        print_type_shared tid in
    (* We output the module that defines the hash function. *)
    Printf.bprintf ob "\
      \nmodule %a = struct\
      \n  type t = %a;;\
      \n  let equal = %a;;\
      \n  let hash = %a;;\
      \nend;;\n"
      longident (hashing_module_ident tid)
      longident type_obj_ident
      print_equal_function_body (equal_function_ident tid)
      print_hash_function_body (hash_function_ident tid);
    (* We output the module that defines the sharing hash table. *)
    Printf.bprintf ob "\
      \nmodule %a = struct\n\
      \n  module Share = Weak.Make (%a);;\n\
      \n  include Share;;\n\
      \n  let find t x = (Obj.magic (Share.find t (Obj.repr x)));;\
      \n  let add t x = Share.add t (Obj.repr x);;\
      \nend;;\n"
      longident (sharing_module_ident tid)
      longident (hashing_module_ident tid);
  end;
  (* We define the sharing hash table. *)
  Printf.bprintf ob "\
    \nlet %a = %a.create %i;;\n"
    longident (sharing_hash_table_ident tid)
    longident (sharing_module_ident tid)
    initial_sharing_hash_table_size;
;;

let generator_underscore (gid, xs) =
  match xs with
    | [] -> Generator (gid, Tuple [])
    | _ -> Generator (gid, Tuple (underscore :: xs))
;;

let genr_pattern_underscore gid root n =
  generator_underscore (gid, genr_args root n)
;;

let generator_info (gname, info, xs) =
  Generator (Lident gname, Tuple (info :: xs))
;;

let genr_pattern_info gname info root n =
  generator_info (gname, info, genr_args root n)
;;

let genr_clause_nullary i gname =
  let genr_constant_pattern gname = make_Constant gname in
  Clause (genr_constant_pattern gname, make_Constant (string_of_int i))
;;

let genr_clauses clause_nullary clause_non_nullary gds default_clauses =
  (* i is the rank of the current nullary generator.
     j is the rank of the current non-nullary generator. *)
  let rec loop (i, j, accu) = function
    | [] -> accu
    | (gname, gts, _rels, _loc) :: gds ->
      match List.length gts with
      | 0 ->
        let accu = clause_nullary i gname accu in
        loop (i + 1, j, accu) gds
      | n ->
        let accu = clause_non_nullary j gname n accu in
        loop (i, j + 1, accu) gds in
  loop (1, 1, default_clauses) gds
;;

(* Generates the hash table internal equality function for (sum) type tname
   (the function named equal_tname in the generated source). *)
let genr_equal_sum tid gds =

  let args = [x; y] in

  let clause_nullary _i _gname accu = accu

  and clause_non_nullary _j gname n accu =
    let xs = genr_args "x" n
    and ys = genr_args "y" n in
    let p1 = generator_underscore (Lident gname, xs)
    and p2 = generator_underscore (Lident gname, ys) in
    let i_xs = genr_numbered_args "x" n in
    let eqs =
      List.map2
        (fun (i, x) y ->
           infix_apply x (get_eq_function_ident_for_arg_i tid gname i) y)
        i_xs ys in
    let body =
      match eqs with
      | eq :: eqs ->
        List.fold_left (fun eq e -> infix_name eq "&&" e) eq eqs
      | _ -> assert false in
    Clause (pair p1 p2, body) :: accu

  and default_clauses = [Clause (underscore, make_Constant "false")] in

  let clauses =
    genr_clauses clause_nullary clause_non_nullary gds default_clauses in
  topletrec [equal_function_ident tid, args, Match (pair x y, clauses)]
;;

let info_pattern h = Record [Lident "hash", make_Constant h]
;;

(* Finds the hash code in the tname values, retrieved in the info fiels. *)
let genr_get_hash_sum tid gds =

  let args = [x] in

  let clause_nullary i gname accu = genr_clause_nullary i gname :: accu

  and clause_non_nullary _j gname n accu =
    Clause
     (genr_pattern_info gname (info_pattern "h") "_x" n, make_Constant "h")
    :: accu

  and default_clauses = [] in

  let clauses =
    genr_clauses clause_nullary clause_non_nullary gds default_clauses in

  topletrec [get_hash_function_ident tid, args, Match (x, clauses)]
;;

(* Generates the hashing function for (sum) type tname
   (the function named hash_tname in the generated code). *)
let genr_hash_sum tid gds =

  let args = [x] in

  let clause_nullary i gname accu = genr_clause_nullary i gname :: accu

  and clause_non_nullary j gname n accu =
    (* The arguments of generator gname. *)
    let i_xis = genr_numbered_args "x" n in
    (* The integer value associated to the generator, its unique rank
       in the type definition.
       This value is added to the sum of all the hash codes of the generator
       arguments. *)
    let chash = make_Constant (string_of_int j) in
    let sum e (i, xi) =
      infix_name
        (apply1 (get_hash_function_ident_for_arg_i tid gname i) xi) "+" e in
    let cl_exp = List.fold_left sum chash i_xis in
    Clause (genr_pattern_underscore (Lident gname) "x" n, cl_exp) :: accu

  and default_clauses = [] in

  let clauses =
    genr_clauses clause_nullary clause_non_nullary gds default_clauses in

  topletrec [hash_function_ident tid, args, Match (x, clauses)]
;;

let genr_construction_function tid gname = function
  | 0 -> assert false
  | n ->
    let args = genr_args "x" n
    and info = make_Constant "info"
    and v = make_Constant "v"
    and hinfo = make_Constant "info.hash" in
    let sharing_module_ident = sharing_module_ident tid
    and tbl = Constant (sharing_hash_table_ident tid) in

    let find_ident = Ldot (sharing_module_ident, "find")
    and add_ident =  Ldot (sharing_module_ident, "add") in

    let with_clause =
      Clause
        (make_Constant "Not_found",
         Sequence (apply2 add_ident tbl v, v, [])) in

    let body =
      Let (Lident "info", [], Record [Lident "hash", make_Constant "0"],
      Let (Lident "v", [], generator_info (gname, info, args),
      Sequence
        (infix_name hinfo "<-" (apply1 (hash_function_ident tid) v),
         Try (apply2 find_ident tbl v, [with_clause]),
         []))) in

    topletrec [mk_sharing_function_ident (Lident gname), args, body]
;;

let genr_construction_functions tid gds =
  List.map
    (fun (gname, gts, _rels, _loc) ->
      genr_construction_function tid gname (List.length gts))
    (List.filter (fun (_s, cts, _rels, _loc) -> List.length cts > 0) gds)
;;

let print_sharing_mli_preamble ob =
  Printf.bprintf ob "\ntype info = private { mutable hash : int };;\n"
;;

let print_sharing_ml_preamble ob =
  Printf.bprintf ob "\ntype info = { mutable hash : int };;\n";
  Printf.bprintf ob "\nlet mk_info h = { hash = h };;\n"
;;

let print_type_private_structure_items ob ntds =
  let print_type_structure_items ob (tname, td) =
    let structure_items = get_structure_items_of_type_declaration td in
    if structure_items <> [] then begin
      Printf.bprintf ob "(* Value defined from type %s *)\n" tname;
      let print si =
	Printf.bprintf ob "%a;;\n\n" Pr_otype.verbatim_structure_item si in
      List.iter print structure_items
    end in
  List.iter (print_type_structure_items ob) ntds
;;

let print_type_private_module ob ntds =
  let private_module_ident =
    match ntds with
    | (tname, _td) :: _ntds -> type_private_module_ident (Lident tname)
    | _ -> assert false in
  Printf.bprintf ob "\nmodule %a = struct\ntype "
    longident private_module_ident;
  let public_ntds =
    List.map
      (fun (tname, td) -> tname,
         add_info_arg_to_type_declaration
           (comment_relations_in_type_declaration
              (make_public_type_declaration td)))
    ntds in
  Pr_otype.type_declarations ob public_ntds;
  print_type_private_structure_items ob ntds;
  Printf.bprintf ob "\nend;;\n";
  (* We now include the just defined private module. *)
  Printf.bprintf ob "\ninclude %a;;\n" longident private_module_ident
;;

let print_ntd ob (tname, td) =
  match td.ptype_kind with
  | Ptype_abstract _rels -> assert false
  | Ptype_variant gds ->
    let tid = Lident tname in
    let teq = genr_equal_sum tid gds in
    let tgh = genr_get_hash_sum tid gds in
    let teh = genr_hash_sum tid gds in
    Pr_code.structure ob [teq; tgh; teh];
    print_hashing_module ob tid;
    Pr_code.structure ob (genr_construction_functions tid gds)
  | Ptype_record _fds -> assert false
;;

let print_sharing ob ntds =
  print_type_private_module ob ntds;
  List.iter (print_ntd ob) ntds
;;
